from aiohttp import ClientSession
from numpy import ndarray
from pandas import DataFrame, RangeIndex, Series
from json import dump
from scipy.sparse import spmatrix
from gooder_ai.types import GetScoreColumnNamesParams, Data
import logging


async def validate_config(validation_url: str, config: dict) -> dict:
    async with ClientSession() as session:
        output = None
        async with session.post(validation_url, json=config) as response:
            if response.status == 200:
                output = await response.json()
            else:
                error = await response.text()
                logging.error(f"Validation failed with {response.status}: {error}")
                raise Exception(f"Validation failed with {response.status}: {error}")
        return output


# TODO: Verify if need to handle multi output classifiers.
def get_score_column_names(params: GetScoreColumnNamesParams) -> list[str]:
    column_names = params["column_names"]
    scores = params["scores"]
    model_name = params["model_name"]

    column_names_count = len(column_names)
    scores_count = len(scores)

    # Log warning if there's a mismatch
    if column_names_count != scores_count:
        logging.warning(
            f"Mismatch: Scores has {scores_count} possible values, but {column_names_count} column names were provided."
        )

    # Ensure output has enough names, filling missing ones
    output = [f"{model_name}_score_{name}" for name in column_names[:scores_count]] + [
        f"{model_name}_score_{column_name}"
        for column_name in scores[column_names_count:]
    ]

    return output


def get_scorer_functions(models_count: int, scorers: list[str]) -> list[str]:
    if models_count != len(scorers):
        logging.warning(
            f"Mismatch:  {models_count} Models, but {len(scorers)} scoring functions were provided."
        )

    output = (
        scorers[:models_count]
        + ["predict_proba" for _ in range(len(scorers), models_count)]
        if models_count != len(scorers)
        else scorers
    )

    return output


# Add column names to dataframe
def transform_dataframe(data: DataFrame, column_names: list[str]) -> DataFrame:
    transformed_dataframe = data.copy()

    # Get existing column names
    default_column_names = transformed_dataframe.columns
    column_count = transformed_dataframe.shape[1]
    final_column_names = []

    if len(column_names) != column_count:
        logging.warning(
            f"Mismatch: DataFrame has {column_count} columns, but {len(column_names)} column names were provided."
        )

    if len(column_names) == column_count:
        # Priority 1: Use provided column names
        final_column_names = column_names
    elif isinstance(default_column_names, RangeIndex) == False:
        # Priority 2: Use existing column names if not RangeIndex
        final_column_names = default_column_names
    else:
        # Priority 3: Generate default names if no names available
        final_column_names = column_names[:column_count] + [
            f"column-{i+1}" for i in range(len(column_names), column_count)
        ]

    # Assign final column names
    transformed_dataframe.columns = final_column_names

    return transformed_dataframe


def transform_unstructured_numpy_array(
    data: ndarray, column_names: list[str]
) -> DataFrame:
    logging.warning("Unstructured numpy array does not have any column name")
    numpy_array = data.copy()

    if numpy_array.ndim > 2:
        raise ValueError("Input data must be a 1D or 2D NumPy array.")

    column_count = (
        1 if numpy_array.ndim == 1 else data.shape[1]
    )  # For 2D arrays, second value of tuple is the column count
    column_names_length = len(column_names)

    if column_names_length != column_count:
        logging.warning(
            f"Mismatch: Numpy array has {column_count} columns, but {column_names_length} column names were provided."
        )

    final_column_names = column_names[:column_count] + [
        f"column-{i+1}" for i in range(column_names_length, column_count)
    ]

    output = DataFrame(numpy_array, columns=final_column_names)
    return output


def transform_structured_numpy_array(
    data: ndarray, column_names: list[str]
) -> DataFrame:
    numpy_array = data.copy()
    default_column_names = list(data.dtype.names)

    if numpy_array.ndim > 2:
        raise ValueError("Input data must be a 1D or 2D NumPy array.")

    column_names_length = len(column_names)
    column_count = 1 if numpy_array.ndim == 1 else data.shape[1]
    if column_names_length != column_count:
        logging.warning(
            f"Mismatch: Numpy array has {column_count} columns, but {column_names_length} column names were provided."
        )

    final_column_names = column_names[: min(column_count, column_names_length)]

    # Use default column names if available
    if len(final_column_names) < column_count:
        remaining_defaults = default_column_names[len(final_column_names) :]
        final_column_names.extend(
            remaining_defaults[: column_count - len(final_column_names)]
        )

    # Generate custom column names for any remaining unnamed columns
    if len(final_column_names) < column_count:
        final_column_names.extend(
            [f"column-{i+1}" for i in range(len(final_column_names), column_count)]
        )

    output = DataFrame(numpy_array, columns=final_column_names)
    return output


def transform_list(
    data: (
        list[str]
        | list[int]
        | list[float]
        | list[list[str]]
        | list[list[int]]
        | list[list[float]]
    ),
    column_names: list[str],
) -> DataFrame:
    logging.warning("List does not have any column name.")
    data_list = data.copy()
    final_column_names = []

    # 1D list handling.
    if isinstance(data_list[0], (str, int, float)):
        data_list = [[value] for value in data_list]
        final_column_names = column_names[:1] if column_names else ["column-1"]

    elif isinstance(data_list[0], list):
        column_count = len(data_list[0])
        if len(column_names) != column_count:
            logging.warning(
                f"Mismatch: DataFrame has {column_count} columns, but {len(column_names)} column names were provided."
            )
        final_column_names = column_names[:column_count] + [
            f"column-{i+1}" for i in range(len(column_names), column_count)
        ]

    else:
        raise ValueError("Input data must be a 1D or 2D list.")

    # Create DataFrame
    output = DataFrame(data_list, columns=final_column_names)
    return output


def get_transformed_data(data: Data, column_names: list[str]) -> DataFrame:
    output = DataFrame()
    if isinstance(data, DataFrame):
        output = transform_dataframe(data, column_names)
    elif isinstance(data, Series):
        output = data.to_frame(name=column_names[0])
    elif isinstance(data, ndarray) and data.dtype.names is None:
        output = transform_unstructured_numpy_array(data, column_names)
    elif isinstance(data, ndarray) and data.dtype.names is not None:
        transform_structured_numpy_array(data, column_names)
    elif isinstance(data, list):
        output = transform_list(data, column_names)
    elif isinstance(data, spmatrix):
        """
        scipy has multiple types of matrix all of them are extending spmatrix
        Ref: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.spmatrix.html#spmatrix
        """
        numpy_array = data.toarray()
        output = transform_unstructured_numpy_array(numpy_array, column_names)

    if isinstance(output.columns, RangeIndex):
        raise ValueError("Invalid columns")
    return output


def save_config(config: dict, file_path: str):
    with open(file_path, "w") as file:
        # dump automatically writes to file
        dump(
            config, file, indent=4
        )  # Ref: https://docs.python.org/3/library/json.html#basic-usage


def get_positive_class(df: DataFrame):
    # Returns the less frequent value in a single-column DataFrame.
    # Rules:
    # - For columns with 1 unique value: returns that value
    # - For binary columns (2 unique values): returns the less frequent value
    # - For empty columns: raises ValueError

    # Args: df: Single-column pandas DataFrame
    # Returns: The value that appears less frequently (or the single value)
    # Raises: ValueError: If input isn't single-column, is empty, or has >2 unique values

    # Validate input structure
    if df.shape[1] != 1:
        logging.error("Input must be a single-column DataFrame")
        raise ValueError("Input must be a single-column DataFrame")

    # Extract values
    if df.empty:  # Get the column name and is it empty or not
        logging.error("DataFrame column is empty")
        raise ValueError("DataFrame column is empty")

    column_name = df.columns[0]
    # Convert to list once
    # Ref: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.value_counts.html
    value_counts = df.value_counts(
        f"{column_name}"
    )  # Column name can be of type string and index depending on the dataFrame
    # Ref: https://pandas.pydata.org/docs/reference/api/pandas.Series.to_dict.html#pandas.Series.to_dict
    count_dictionary = value_counts.to_dict()

    if len(count_dictionary.keys()) > 2:
        logging.error(
            f"Column must contain maximum 2 unique values (found {len(count_dictionary.keys())})"
        )
        raise ValueError(
            f"Column must contain maximum 2 unique values (found {len(count_dictionary.keys())})"
        )

    least_frequent = min(count_dictionary.keys(), key=lambda k: count_dictionary[k])
    return least_frequent



from aiohttp import ClientSession
from gooder_ai.view.types import ExecuteGraphQLParams


Create_Shared_View = """
mutation CreateSharedView(
    $input: CreateSharedViewInput!
    $condition: ModelSharedViewConditionInput
) {
    createSharedView(input: $input, condition: $condition) {
        id
        datasetPath
        configPath
        owner
        accessibleTo
        createdAt
        updatedAt
        __typename
    }
}"""

Update_Shared_View = """
mutation UpdateSharedView(
    $input: UpdateSharedViewInput!
    $condition: ModelSharedViewConditionInput
) {
    updateSharedView(input: $input, condition: $condition) {
        id
        datasetPath
        configPath
        owner
        accessibleTo
        createdAt
        updatedAt
        __typename
    }
}"""


async def execute_graphql_query(input: ExecuteGraphQLParams) -> dict:
    token = input["token"]
    api_url = input["api_url"]
    variables = input["variables"]
    mutation_type = input["mutation"]
    query = (
        Create_Shared_View
        if mutation_type == "createSharedView"
        else Update_Shared_View
    )

    headers = {"Content-Type": "application/json", "Authorization": f"Bearer {token}"}

    async with ClientSession() as session:
        output = None
        async with session.post(
            api_url, json={"query": query, "variables": variables}, headers=headers
        ) as response:
            output = await response.json()
            if (output["data"][mutation_type]) is None:
                raise Exception(f"GraphQL query failed: {output}")
        return output



from typing import TypedDict


class AuthenticateParams(TypedDict):
    email: str
    password: str
    app_client_id: str
    identity_pool_id: str
    user_pool_id: str


class AuthenticationResult(TypedDict):
    AccessToken: str
    IdToken: str
    RefreshToken: str
    ExpiresIn: int
    TokenType: str


class CognitoClientResponse(TypedDict):
    AuthenticationResult: AuthenticationResult


class CognitoIdentityResponse(TypedDict):
    IdentityId: str


class Credentials(TypedDict):
    AccessKeyId: str
    SecretKey: str
    SessionToken: str
    Expiration: str


class CognitoCredentialsResponse(TypedDict):
    IdentityId: str
    Credentials: Credentials


class AuthenticateReturn(TypedDict):
    cognito_client_response: CognitoClientResponse
    cognito_identity_response: CognitoIdentityResponse
    cognito_credentials: CognitoCredentialsResponse



from typing import TypedDict, Literal
from botocore.client import BaseClient
from pandas import DataFrame
from gooder_ai.auth.types import AuthenticateReturn


class UploadParams(TypedDict):
    s3_client: BaseClient  # Ref: https://stackoverflow.com/a/68006865
    identity_id: str
    file_name: str
    mode: Literal["public", "protected", "private"]
    bucket_name: str
    data: DataFrame | dict


class UploadFileParams(TypedDict):
    credentials: AuthenticateReturn

    data: DataFrame
    file_name: str
    config: dict
    mode: Literal["public", "protected", "private"]
    bucket_name: str

    upload_data_to_gooder: bool
    upload_config_to_gooder: bool



from gooder_ai.types import AmplifyEnv

# The following configuration values are environment-specific and need to be updated when switching to a new environment.
# For example, we are currently on environment "seventeen," and when transitioning to environment "eighteen,"
# these values should be updated accordingly.
Amplify_Env: dict[str, AmplifyEnv] = {
    "latest": {
        "API_URL": "https://y5u2lvnrl5flblw45ojgyckxau.appsync-api.us-east-1.amazonaws.com/graphql",  # GraphQL API URL
        "App_Client_ID": "5ml0oe8lbn6e9ail331gmg11n9",  # App_Client_ID is the client ID for an app in a Cognito User Pool
        "Identity_Pool_ID": "us-east-1:60a5f15a-7736-4d81-9f7c-c2306deed47f",  # Identity_Pool_ID is the ID for an Amazon Cognito Identity Pool
        "User_Pool_ID": "us-east-1_M38tSlB29",  # UserPool_ID is the ID for the Cognito User Pool that manages user authentication
        "Bucket_Name": "mlviz1049e7ed428840c3bb9773b134209eddb4289-eighteen",  # S3 bucket name storing datasets and configs
        "Validation_API_URL": "https://l3tgixzm39.execute-api.us-east-1.amazonaws.com/eighteen/validate",  # Validation API URL
        "Base_URL": "https://latest.gooder.ai",  # This is the URL for gooder ai application
        "AWS_Region_Name": "us-east-1",
    },
    "seventeen": {
        "API_URL": "https://ojykj5ys3rflbprrjwiyi524oy.appsync-api.us-east-1.amazonaws.com/graphql",  # GraphQL API URL
        "App_Client_ID": "6hmbs2hm1dbpantdbpfnokina",  # App_Client_ID is the client ID for an app in a Cognito User Pool
        "Identity_Pool_ID": "us-east-1:9682a196-3376-4d36-b5d8-d46862f47d7b",  # Identity_Pool_ID is the ID for an Amazon Cognito Identity Pool
        "User_Pool_ID": "us-east-1_wzrj8LRkh",  # UserPool_ID is the ID for the Cognito User Pool that manages user authentication
        "Bucket_Name": "mlviz1049e7ed428840c3bb9773b134209edd111800-seventeen",  # S3 bucket name storing datasets and configs
        "Validation_API_URL": "https://htknl8tmxi.execute-api.us-east-1.amazonaws.com/seventeen/validate",  # Validation API URL
        "Base_URL": "https://app.gooder.ai",  # This is the URL for gooder ai application
        "AWS_Region_Name": "us-east-1",
    },
}



from typing import TypedDict, Literal


class MutationVariables(TypedDict):
    datasetPath: str | None
    configPath: str | None
    id: str


class ExecuteGraphQLParams(TypedDict):
    mutation: Literal["createSharedView", "updateSharedView"]
    variables: dict[Literal["input"], MutationVariables]
    token: str
    api_url: str



from datetime import datetime
from uuid import uuid4
from webbrowser import open_new
from gooder_ai.globals import Amplify_Env
from gooder_ai.s3 import upload_files
from gooder_ai.view import execute_graphql_query, ExecuteGraphQLParams
from gooder_ai.auth import authenticate, AuthenticateReturn
from gooder_ai.utils import (
    validate_config,
    get_transformed_data,
    get_score_column_names,
    get_scorer_functions,
    save_config,
    get_positive_class,
)
from gooder_ai.types import (
    ViewMeta,
    ValuateModelOutput,
    ScikitModel,
    Credentials,
    AWSVariables,
    ColumnNames,
    Data,
)
from asyncio import to_thread
from pandas import DataFrame, concat
import logging
import os


async def valuate_model(
    models: list[ScikitModel],
    x_data: Data,
    y: Data,
    config: dict,
    view_meta: ViewMeta,
    auth_credentials: Credentials={},
    model_names: list[str]=[],
    scorer_names: list[str] = [],
    column_names: ColumnNames = {},
    included_columns: list[str] = [],
    aws_variables: AWSVariables = {},
    upload_data_to_gooder: bool = True,
    upload_config_to_gooder: bool = True,
    amplify_env: str = "seventeen",  # For production the amplify version is seventeen
    max_size_uploaded_data: int = 10,
    max_size_saved_data: int = 1000,
) -> ValuateModelOutput:
    logging.info("Model valuation started.")
    email = auth_credentials.get("email", None)
    password = auth_credentials.get("password", None)
    mode = view_meta.get("mode", "private")
    view_id = view_meta.get("view_id", None)
    dataset_name = view_meta.get("dataset_name", f"{datetime.now().timestamp()}")
    # Create a copy of the config to avoid modifying the original
    raw_config = {**config, "scores": []}  # Initialize scores to an empty list

    # AWS Global Variables
    api_url = aws_variables.get("api_url", Amplify_Env[amplify_env]["API_URL"])
    app_client_id = aws_variables.get(
        "app_client_id", Amplify_Env[amplify_env]["App_Client_ID"]
    )
    identity_pool_id = aws_variables.get(
        "identity_pool_id", Amplify_Env[amplify_env]["Identity_Pool_ID"]
    )
    user_pool_id = aws_variables.get(
        "user_pool_id", Amplify_Env[amplify_env]["User_Pool_ID"]
    )
    bucket_name = aws_variables.get(
        "bucket_name", Amplify_Env[amplify_env]["Bucket_Name"]
    )
    base_url = aws_variables.get("base_url", Amplify_Env[amplify_env]["Base_URL"])
    validation_api_url = aws_variables.get(
        "validation_api_url", Amplify_Env[amplify_env]["Validation_API_URL"]
    )

    transformed_x_data = get_transformed_data(
        x_data, column_names.get("dataset_column_names", [])
    )

    filtered_x_data = (
        transformed_x_data.filter(items=included_columns)
        if len(included_columns) > 0
        else transformed_x_data
    )

    transformed_y_data = get_transformed_data(
        y, [column_names.get("dependent_variable_name", "dependent_variable")]
    )

    combined_dataframe = concat([filtered_x_data, transformed_y_data], axis=1)

    scorer_functions = get_scorer_functions(len(models), scorer_names)

    for index, model in enumerate(models):
        scorer = getattr(model, scorer_functions[index])
        # Ref: https://stackoverflow.com/questions/52763325/how-to-obtain-only-the-name-of-a-models-object-in-scikitlearn

        # Get model name - use from model_names if provided, otherwise use class name
        model_name = (
            model_names[index]
            if index < len(model_names)
            else f"{type(model).__name__}{index}"
        )

        if scorer is None:
            logging.error(
                f"Failed: Input model instance doesn't have a '{scorer_functions[index]}' method to score the model performance."
            )
            raise Exception(
                f"Failed: Input model instance doesn't have a '{scorer_functions[index]}' method to score the model performance."
            )
        model_classes = getattr(model, "classes_", None)
        if model_classes is None:
            logging.error(
                "Failed: Input model instance doesn't have classes to classify the target variable"
            )
            raise Exception(
                "Failed: Input model instance doesn't have classes to classify the target variable."
            )
        score_column_names = get_score_column_names(
            {
                "column_names": column_names.get("score_column_names", []),
                "scores": model_classes,
                "model_name": model_name,
            }
        )

        full_scores_df = DataFrame(
            scorer(x_data), columns=score_column_names, index=combined_dataframe.index
        )

        if len(model_classes) == 2:
            # handle binary classification
            positive_class = get_positive_class(transformed_y_data)
            positive_class_index = list(model_classes).index(positive_class)
            new_score_entry = {"fieldName": f"{model_name}_score"}
            logging.info(f"Creating new score entry: {new_score_entry}")

            positive_class_score_name = score_column_names[positive_class_index]
            full_scores_df = full_scores_df[[positive_class_score_name]].rename(
                columns={positive_class_score_name: f"{model_name}_score"}
            )

            raw_config["scores"] = raw_config.get("scores", []) + [new_score_entry]
            logging.info(
                f"Updated config scores in binary classification: {raw_config['scores']}"
            )
        else:
            # handle multi-class classification
            raw_config["scores"] = raw_config.get("scores", []) + score_column_names
            logging.info(
                f"Updated config scores in multi-class classification: {raw_config['scores']}"
            )

        combined_dataframe = concat(
            [
                combined_dataframe,
                full_scores_df,
            ],
            axis=1,
        )

    # Ref: https://stackoverflow.com/questions/18089667/how-to-estimate-how-much-memory-a-pandas-dataframe-will-need
    # Ref: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.memory_usage.html#
    df_memory = combined_dataframe.memory_usage(deep=True).sum() / (
        1024 * 1024
    )  # Recalculate the DataFrame size in MB after performing calculations

    if upload_data_to_gooder == True and df_memory >= max_size_uploaded_data:
        logging.error(
            "Failed: Provided DataFrame size exceeding the maximum upload limit."
        )
        raise Exception(
            "Provided DataFrame size is exceeding the maximum upload limit. Please reduce the size of the DataFrame."
        )

    elif upload_data_to_gooder == False and df_memory >= max_size_saved_data:
        logging.error(
            "Failed: Provided DataFrame size exceeding the maximum save data limit."
        )
        raise Exception(
            "Provided DataFrame size is exceeding the maximum save data limit. Please reduce the size of the DataFrame."
        )

    logging.info("Started: Validating config as per the Gooder AI schema.")
    parsed_config = await validate_config(validation_api_url, raw_config)

    if parsed_config["success"] == False:
        logging.error("Failed: Validating config as per the Gooder AI schema.")
        raise Exception("Invalid configuration", parsed_config["error"])
    else:
        logging.info("Success: Validating config as per the Gooder AI schema.")

    logging.info("Started: Authenticating for the Gooder AI platform.")

    credentials: AuthenticateReturn | None = (
        None  # TODO: update return type in next PR.
    )
    if upload_data_to_gooder == True or upload_config_to_gooder == True:
        if isinstance(email, str) and isinstance(password, str):
            credentials = authenticate(
                {
                    "email": email,
                    "password": password,
                    "app_client_id": app_client_id,
                    "identity_pool_id": identity_pool_id,
                    "user_pool_id": user_pool_id,
                }
            )
            logging.info("Success: Authenticating for the Gooder AI platform.")
        else:
            logging.error("Failed: Email and password are required for authentication.")
            raise Exception("Email and password are required for authentication.")

    parsed_config["data"][
        "datasetID"
    ] = f"{dataset_name}.csv/Sheet1"  # override datasetID of config to match with dataset.

    if upload_data_to_gooder == True:
        logging.info("Started: Uploading dataset to the Gooder AI platform.")

    if upload_config_to_gooder == True:
        logging.info("Started: Uploading config to the Gooder AI platform.")

    path_dictionary = {
        "csv_path": None,
        "config_path": None,
    }

    if credentials is not None:
        path_dictionary = await upload_files(
            {
                "credentials": credentials,
                "data": combined_dataframe,
                "config": parsed_config["data"],
                "file_name": dataset_name,
                "mode": mode,
                "bucket_name": bucket_name,
                "upload_data_to_gooder": upload_data_to_gooder,
                "upload_config_to_gooder": upload_config_to_gooder,
            }
        )

    csv_path = path_dictionary["csv_path"]
    config_path = path_dictionary["config_path"]

    if csv_path is None and upload_data_to_gooder == True:
        logging.error("Failed: Uploading dataset to the Gooder AI platform.")
        raise Exception("Failed to upload dataset")
    elif upload_data_to_gooder == False:
        local_dataset_filename = f"{dataset_name}.csv"
        # Ref: https://docs.python.org/3/library/asyncio-task.html#running-in-threads
        # TODO: Do we need to_thread ?
        await to_thread(
            lambda: combined_dataframe.to_csv(local_dataset_filename, index=False)
        )
        logging.info(f"Dataset saved as {os.path.abspath(local_dataset_filename)}.")

    if config_path is None and upload_config_to_gooder == True:
        logging.error("Failed: Uploading config to the Gooder AI platform.")
        raise Exception("Failed to upload config")
    elif upload_config_to_gooder == False:
        local_config_filename = (
            f"{dataset_name}.json"  # JSON file name is same as dataset_name file name
        )
        # Ref: https://docs.python.org/3/library/asyncio-task.html#running-in-threads
        # TODO: Do we need to_thread ?
        await to_thread(
            lambda: save_config(parsed_config["data"], local_config_filename)
        )
        logging.info(f"Config saved as {os.path.abspath(local_config_filename)}.")

    mutation_type = (
        "updateSharedView" if isinstance(view_id, str) else "createSharedView"
    )

    id: str | None = None
    if credentials is not None:
        view_params: ExecuteGraphQLParams = {
            "api_url": api_url,
            "token": credentials["cognito_client_response"]["AuthenticationResult"][
                "AccessToken"
            ],
            "mutation": mutation_type,
            "variables": {
                "input": {
                    "configPath": config_path,
                    "datasetPath": csv_path,
                    "id": view_id if isinstance(view_id, str) else f"{uuid4()}",
                }
            },
        }
        view = await execute_graphql_query(view_params)
        id = view["data"][mutation_type]["id"]
        message = (
            f"View with ID {id} has been successfully updated using the provided view ID: {view_id}."
            if mutation_type == "updateSharedView"
            else f"A new view has been created successfully. Your view ID is {id}. Please save it for future reference and reuse."
        )
        logging.info(message)
        logging.info("Model valuation can be continued on the Gooder AI platform now.")

    view_url = base_url if id is None else f"{base_url}?view={id}"
    open_new(view_url)
    return {"view_id": id, "view_url": view_url}



import boto3
from pandas import DataFrame
from io import BytesIO
from json import dumps
from asyncio import TaskGroup
from botocore.exceptions import ClientError
from gooder_ai.s3.types import UploadParams, UploadFileParams
from botocore.client import BaseClient
from typing import cast


# Convert dataframe to an in-memory csv sheet Ref: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_csv.html#pandas.DataFrame.to_csv
# Using buffer reference: https://stackoverflow.com/a/40540730
# Using BytesIO instead of StringIO because, s3 upload file function uses BytesIO
async def transform_dataframe_to_csv_file_object(data: DataFrame) -> BytesIO:
    csv_buffer = BytesIO()
    data.to_csv(csv_buffer, index=False)
    csv_buffer.seek(
        0
    )  # When we write something in buffer, file pointer is moved at the end of file. Here we are resetting the file pointer position to start after writing to file. Ref: https://stackoverflow.com/a/53485819
    return csv_buffer


# Convert config dictionary to a string to an in-memory json Ref: https://docs.python.org/3/library/json.html#basic-usage
# Using buffer reference: https://docs.python.org/3/library/io.html#binary-i-o
# Using BytesIO instead of StringIO because, s3 upload file function uses BytesIO
async def transform_config_to_json_file_object(config: dict) -> BytesIO:
    json_string = dumps(config)
    json_buffer = BytesIO(json_string.encode("utf-8"))
    json_buffer.seek(
        0
    )  # When we write something in buffer, file pointer is moved at the end of file. Here we are resetting the file pointer position to start after writing to file. Ref: https://stackoverflow.com/a/53485819
    return json_buffer


async def upload_handler(input: UploadParams) -> str | None:
    s3_client = input["s3_client"]
    identity_id = input["identity_id"]
    data = input["data"]
    mode = input["mode"]
    bucket_name = input["bucket_name"]
    file_name = input["file_name"]

    path: str | None = None
    parsed_data: BytesIO | None = None

    if isinstance(data, DataFrame):
        # The path to upload file is referenced from amplify. Ref: https://docs.amplify.aws/gen1/javascript/build-a-backend/storage/path/#using-protected-accesslevel
        path = f"{mode}/{identity_id}/data/{file_name}.csv"
        parsed_data = await transform_dataframe_to_csv_file_object(data)
    else:
        path = f"{mode}/{identity_id}/config/{file_name}.json"
        parsed_data = await transform_config_to_json_file_object(data)

    try:
        s3_client.upload_fileobj(parsed_data, bucket_name, path)
    except ClientError as e:
        print(f"Error uploading file: {e}")
        path = None
    return path


async def upload_files(input: UploadFileParams) -> dict[str, str | None]:
    credentials = input["credentials"]
    data = input["data"]
    file_name = input["file_name"]
    config = input["config"]
    mode = input["mode"]
    bucket_name = input["bucket_name"]
    upload_data_to_gooder = input["upload_data_to_gooder"]
    upload_config_to_gooder = input["upload_config_to_gooder"]
    aws_access_key_id = credentials["cognito_credentials"]["Credentials"]["AccessKeyId"]
    aws_secret_access_key = credentials["cognito_credentials"]["Credentials"][
        "SecretKey"
    ]
    aws_session_token = credentials["cognito_credentials"]["Credentials"][
        "SessionToken"
    ]
    identity_id = credentials["cognito_credentials"]["IdentityId"]

    # Reference for using temporary credentials : https://docs.aws.amazon.com/IAM/latest/UserGuide/id_credentials_temp_use-resources.html
    s3_client = cast(
        BaseClient,
        boto3.client(
            "s3",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
        ),
    )  # Mimicing "as" assertion to prevent type mismatch

    csv_path: str | None = None
    config_path: str | None = None

    """
    A new alternative to create and run tasks concurrently and wait for their completion 
    is asyncio.TaskGroup. TaskGroup provides stronger safety guarantees 
    than gather for scheduling a nesting of subtasks: if a task (or a subtask,
    a task scheduled by a task) raises an exception, TaskGroup will, while
    gather will not, cancel the remaining scheduled tasks).
    """
    csv_task = None
    config_task = None

    async with TaskGroup() as tg:
        if upload_data_to_gooder == True:
            csv_task = tg.create_task(
                upload_handler(
                    {
                        "identity_id": identity_id,
                        "s3_client": s3_client,
                        "data": data,
                        "file_name": file_name,
                        "mode": mode,
                        "bucket_name": bucket_name,
                    }
                )
            )

        if upload_config_to_gooder == True:
            config_task = tg.create_task(
                upload_handler(
                    {
                        "bucket_name": bucket_name,
                        "file_name": file_name,
                        "data": config,
                        "identity_id": identity_id,
                        "mode": mode,
                        "s3_client": s3_client,
                    }
                )
            )

    if csv_task is not None:
        csv_path = csv_task.result()

    if config_task is not None:
        config_path = config_task.result()

    return {"csv_path": csv_path, "config_path": config_path}
